@@ -5,22 +5,25 @@
 # SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
 # SPDX-FileCopyrightText: 2021 Alvar Penning
 # SPDX-FileCopyrightText: 2021 Alliander N.V. <https://alliander.com>
+# SPDX-FileCopyrightText: 2021 Robin Vobruba <hoijui.quaero@gmail.com>
+# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
 #
 # SPDX-License-Identifier: GPL-3.0-or-later
 
 """Functions for manipulating the comment headers of files."""
 
-# pylint: disable=too-many-arguments
 
+import argparse
 import datetime
 import logging
+import os
 import re
 import sys
 from argparse import ArgumentParser
 from gettext import gettext as _
 from os import PathLike
 from pathlib import Path
-from typing import List, NamedTuple, Optional, Sequence
+from typing import Iterable, List, NamedTuple, Optional, Sequence, Tuple
 
 from binaryornot.check import is_binary
 from boolean.boolean import ParseError
@@ -33,10 +36,12 @@ from ._comment import (
     EXTENSION_COMMENT_STYLE_MAP_LOWERCASE,
     FILENAME_COMMENT_STYLE_MAP_LOWERCASE,
     NAME_STYLE_MAP,
+    CCommentStyle,
     CommentCreateError,
     CommentParseError,
     CommentStyle,
     EmptyCommentStyle,
+    HtmlCommentStyle,
     PythonCommentStyle,
     UncommentableCommentStyle,
 )
@@ -49,21 +54,20 @@ from ._util import (
     detect_line_endings,
     extract_spdx_info,
     make_copyright_line,
+    merge_copyright_lines,
     spdx_identifier,
 )
 from .project import Project
 
 _LOGGER = logging.getLogger(__name__)
 
-_ENV = Environment(
-    loader=PackageLoader("reuse", "templates"), trim_blocks=True
-)
+_ENV = Environment(loader=PackageLoader("reuse", "templates"), trim_blocks=True)
 DEFAULT_TEMPLATE = _ENV.get_template("default_template.jinja2")
 
 _NEWLINE_PATTERN = re.compile(r"\n", re.MULTILINE)
 
 
-class _TextSections(NamedTuple):  # pylint: disable=too-few-public-methods
+class _TextSections(NamedTuple):
     """Used to split up text in three parts."""
 
     before: str
@@ -97,10 +101,10 @@ def _create_new_header(
     rendered = template.render(
         copyright_lines=sorted(spdx_info.copyright_lines),
         spdx_expressions=sorted(map(str, spdx_info.spdx_expressions)),
-    )
+    ).strip("\n")
 
     if template_is_commented:
-        result = rendered.strip("\n")
+        result = rendered
     else:
         result = style.create_comment(rendered, force_multi=force_multi).strip(
             "\n"
@@ -124,6 +128,7 @@ def _create_new_header(
     return result
 
 
+# pylint: disable=too-many-arguments
 def create_header(
     spdx_info: SpdxInfo,
     header: str = None,
@@ -131,6 +136,7 @@ def create_header(
     template_is_commented: bool = False,
     style: CommentStyle = None,
     force_multi: bool = False,
+    merge_copyrights: bool = False,
 ) -> str:
     """Create a header containing *spdx_info*. *header* is an optional argument
     containing a header which should be modified to include *spdx_info*. If
@@ -157,10 +163,19 @@ def create_header(
                 "existing header contains an erroneous SPDX expression"
             ) from err
 
+        if merge_copyrights:
+            spdx_copyrights = merge_copyright_lines(
+                spdx_info.copyright_lines.union(existing_spdx.copyright_lines),
+            )
+        else:
+            spdx_copyrights = spdx_info.copyright_lines.union(
+                existing_spdx.copyright_lines
+            )
+
         # TODO: This behaviour does not match the docstring.
         spdx_info = SpdxInfo(
             spdx_info.spdx_expressions.union(existing_spdx.spdx_expressions),
-            spdx_info.copyright_lines.union(existing_spdx.copyright_lines),
+            spdx_copyrights,
         )
 
     new_header += _create_new_header(
@@ -170,7 +185,7 @@ def create_header(
         style=style,
         force_multi=force_multi,
     )
-    return new_header + "\n"
+    return new_header
 
 
 def _indices_of_newlines(text: str) -> Sequence[int]:
@@ -214,6 +229,22 @@ def _find_first_spdx_comment(
     raise MissingSpdxInfo()
 
 
+def _extract_shebang(prefix: str, text: str) -> Tuple[str, str]:
+    """Remove all lines that start with the shebang prefix from *text*. Return a
+    tuple of (shebang, reduced_text).
+    """
+    shebang_lines = []
+    for line in text.splitlines():
+        if line.startswith(prefix):
+            shebang_lines.append(line)
+            text = text.replace(line, "", 1)
+        else:
+            shebang = "\n".join(shebang_lines)
+            break
+    return (shebang, text)
+
+
+# pylint: disable=too-many-arguments
 def find_and_replace_header(
     text: str,
     spdx_info: SpdxInfo,
@@ -221,6 +252,7 @@ def find_and_replace_header(
     template_is_commented: bool = False,
     style: CommentStyle = None,
     force_multi: bool = False,
+    merge_copyrights: bool = False,
 ) -> str:
     """Find the first SPDX comment block in *text*. That comment block is
     replaced by a new comment block containing *spdx_info*. It is formatted as
@@ -241,7 +273,6 @@ def find_and_replace_header(
     :raises MissingSpdxInfo: if the generated comment is missing SPDX
         information.
     """
-    # pylint: disable=too-many-branches
     if style is None:
         style = PythonCommentStyle
 
@@ -254,28 +285,27 @@ def find_and_replace_header(
     if style is EmptyCommentStyle:
         after = ""
 
-    # pylint: disable=logging-format-interpolation
     _LOGGER.debug(f"before = {repr(before)}")
     _LOGGER.debug(f"header = {repr(header)}")
     _LOGGER.debug(f"after = {repr(after)}")
 
-    # Extract shebang from header and put it in before. It's a bit messy, but
-    # it ends up working.
-    if header.startswith("#!") and not before.strip():
-        before = ""
-        for line in header.splitlines():
-            if line.startswith("#!"):
-                before = before + "\n" + line
-                header = header.replace(line, "", 1)
-            else:
-                break
-    elif after.startswith("#!") and not any((before, header)):
-        for line in after.splitlines():
-            if line.startswith("#!"):
-                before = before + "\n" + line
-                after = after.replace(line, "", 1)
-            else:
-                break
+    # Keep special first-line-of-file lines as the first line in the file,
+    # or say, move our comments after it.
+    for prefix in (
+        prefix
+        for com_style, prefix in (
+            (CCommentStyle, "#!"),  # e.g. V-Lang
+            (HtmlCommentStyle, "<?xml"),  # e.g. XML/XHTML
+            (PythonCommentStyle, "#!"),  # e.g. Shell, Python
+        )
+        if style is com_style
+    ):
+        # Extract shebang from header and put it in before. It's a bit messy,
+        # but it ends up working.
+        if header.startswith(prefix) and not before.strip():
+            before, header = _extract_shebang(prefix, header)
+        elif after.startswith(prefix) and not any((before, header)):
+            before, after = _extract_shebang(prefix, after)
 
     header = create_header(
         spdx_info,
@@ -284,13 +314,14 @@ def find_and_replace_header(
         template_is_commented=template_is_commented,
         style=style,
         force_multi=force_multi,
+        merge_copyrights=merge_copyrights,
     )
 
-    new_text = header.strip("\n")
+    new_text = f"{header}\n"
     if before.strip():
-        new_text = before.strip("\n") + "\n\n" + new_text
+        new_text = f"{before.rstrip()}\n\n{new_text}"
     if after.strip():
-        new_text = new_text + "\n\n" + after.lstrip("\n")
+        new_text = f"{new_text}\n{after.lstrip()}"
     return new_text
 
 
@@ -340,19 +371,27 @@ def _verify_paths_line_handling(
 
 
 def _verify_paths_comment_style(paths: List[Path], parser: ArgumentParser):
+    unrecognised_files = []
+
     for path in paths:
         style = _get_comment_style(path)
         not_uncommentable = not _is_uncommentable(path)
 
         # TODO: This check is duplicated.
         if style is None and not_uncommentable:
-            parser.error(
+            unrecognised_files.append(path)
+
+    if unrecognised_files:
+        parser.error(
+            "{}\n{}".format(
                 _(
-                    "'{path}' does not have a recognised file extension,"
-                    " please use --style, --explicit-license or"
-                    " --skip-unrecognised"
-                ).format(path=path)
+                    "The following files do not have a recognised file"
+                    " extension. Please use --style, --force-dot-license or"
+                    " --skip-unrecognised:"
+                ),
+                "\n".join(str(path) for path in unrecognised_files),
             )
+        )
 
 
 def _find_template(project: Project, name: str) -> Template:
@@ -386,6 +425,8 @@ def _add_header_to_file(
     template_is_commented: bool,
     style: Optional[str],
     force_multi: bool = False,
+    skip_existing: bool = False,
+    merge_copyrights: bool = False,
     out=sys.stdout,
 ) -> int:
     """Helper function."""
@@ -403,6 +444,17 @@ def _add_header_to_file(
     with path.open("r", encoding="utf-8", newline="") as fp:
         text = fp.read()
 
+    # Ideally, this check is done elsewhere. But that would necessitate reading
+    # the file contents before this function is called.
+    if skip_existing and contains_spdx_info(text):
+        out.write(
+            _(
+                "Skipped file '{path}' already containing SPDX information"
+            ).format(path=path)
+        )
+        out.write("\n")
+        return result
+
     # Detect and remember line endings for later conversion.
     line_ending = detect_line_endings(text)
     # Normalise line endings.
@@ -416,6 +468,7 @@ def _add_header_to_file(
             template_is_commented=template_is_commented,
             style=style,
             force_multi=force_multi,
+            merge_copyrights=merge_copyrights,
         )
     except CommentCreateError:
         out.write(
@@ -443,6 +496,16 @@ def _add_header_to_file(
     return result
 
 
+def _verify_write_access(paths: Iterable[PathLike], parser: ArgumentParser):
+    not_writeable = [
+        str(path) for path in paths if not os.access(path, os.W_OK)
+    ]
+    if not_writeable:
+        parser.error(
+            _("can't write to '{}'").format("', '".join(not_writeable))
+        )
+
+
 def add_arguments(parser) -> None:
     """Add arguments to parser."""
     parser.add_argument(
@@ -462,7 +525,7 @@ def add_arguments(parser) -> None:
     parser.add_argument(
         "--year",
         "-y",
-        action="store",
+        action="append",
         type=str,
         help=_("year of copyright statement, optional"),
     )
@@ -492,6 +555,11 @@ def add_arguments(parser) -> None:
         action="store_true",
         help=_("do not include year in statement"),
     )
+    parser.add_argument(
+        "--merge-copyrights",
+        action="store_true",
+        help=_("merge copyright lines if copyright statements are identical"),
+    )
     parser.add_argument(
         "--single-line",
         action="store_true",
@@ -505,19 +573,37 @@ def add_arguments(parser) -> None:
     parser.add_argument(
         "--explicit-license",
         action="store_true",
-        help=_("place header in path.license instead of path"),
+        help=argparse.SUPPRESS,
+    )
+    parser.add_argument(
+        "--force-dot-license",
+        action="store_true",
+        help=_("write a .license file instead of a header inside the file"),
+    )
+    parser.add_argument(
+        "--recursive",
+        "-r",
+        action="store_true",
+        help=_(
+            "add headers to all files under specified directories recursively"
+        ),
     )
     parser.add_argument(
         "--skip-unrecognised",
         action="store_true",
         help=_("skip files with unrecognised comment styles"),
     )
-    parser.add_argument("path", action="store", nargs="+", type=PathType("r+"))
+    parser.add_argument(
+        "--skip-existing",
+        action="store_true",
+        help=_("skip files that already contain SPDX information"),
+    )
+    parser.add_argument("path", action="store", nargs="+", type=PathType("r"))
 
 
 def run(args, project: Project, out=sys.stdout) -> int:
     """Add headers to files."""
-    # pylint: disable=too-many-branches
+    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
     if not any((args.copyright, args.license)):
         args.parser.error(_("option --copyright or --license is required"))
 
@@ -538,11 +624,37 @@ def run(args, project: Project, out=sys.stdout) -> int:
                 " --style"
             )
         )
+    if args.explicit_license:
+        _LOGGER.warning(
+            _(
+                "--explicit-license has been deprecated in favour of"
+                " --force-dot-license"
+            )
+        )
+        args.force_dot_license = True
 
-    paths = [_determine_license_path(path) for path in args.path]
+    if args.recursive:
+        paths = set()
+        all_files = [path.resolve() for path in project.all_files()]
+        for path in args.path:
+            if path.is_file():
+                paths.add(path)
+            else:
+                paths |= {
+                    child
+                    for child in all_files
+                    if path.resolve() in child.parents
+                }
+    else:
+        paths = args.path
+
+    paths = {_determine_license_path(path) for path in paths}
+
+    if not args.force_dot_license:
+        _verify_write_access(paths, args.parser)
 
     # Verify line handling and comment styles before proceeding
-    if args.style is None and not args.explicit_license:
+    if args.style is None and not args.force_dot_license:
         _verify_paths_line_handling(
             paths,
             args.parser,
@@ -569,8 +681,10 @@ def run(args, project: Project, out=sys.stdout) -> int:
 
     year = None
     if not args.exclude_year:
-        if args.year:
-            year = args.year
+        if args.year and len(args.year) > 1:
+            year = f"{min(args.year)} - {max(args.year)}"
+        elif args.year:
+            year = args.year.pop()
         else:
             year = datetime.date.today().year
 
@@ -592,7 +706,7 @@ def run(args, project: Project, out=sys.stdout) -> int:
     result = 0
     for path in paths:
         uncommentable = _is_uncommentable(path)
-        if uncommentable or args.explicit_license:
+        if uncommentable or args.force_dot_license:
             new_path = _determine_license_suffix_path(path)
             if uncommentable:
                 _LOGGER.info(
@@ -604,13 +718,15 @@ def run(args, project: Project, out=sys.stdout) -> int:
             path = Path(new_path)
             path.touch()
         result += _add_header_to_file(
-            path,
-            spdx_info,
-            template,
-            commented,
-            args.style,
-            args.multi_line,
-            out,
+            path=path,
+            spdx_info=spdx_info,
+            template=template,
+            template_is_commented=commented,
+            style=args.style,
+            force_multi=args.multi_line,
+            skip_existing=args.skip_existing,
+            merge_copyrights=args.merge_copyrights,
+            out=out,
         )
 
     return min(result, 1)
